///////////////////////////////////////
//        Data Master 2003           //
//   Copyright (c) 1993-2003 RRR     //
///////////////////////////////////////

unit AxisDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ImgList, StdCtrls, Buttons, Spin, ExtCtrls, ColorGrd, DMPlot,
  DMFloatEdit, DMComboBox;

type
  TAxisPropsForm = class(TForm)
    OkBitBtn: TBitBtn;
    CancelBitBtn: TBitBtn;
    HelpBitBtn: TBitBtn;
    ImageList: TImageList;
    FontDialog: TFontDialog;
    ColorDialog: TColorDialog;
    PageControl: TPageControl;
    ScaleTabSheet: TTabSheet;
    LabelsTabSheet: TTabSheet;
    LineTabSheet: TTabSheet;
    ManualScalePanel: TPanel;
    ManualScaleRadioButton: TRadioButton;
    MinScaleLabel: TLabel;
    MaxScaleLabel: TLabel;
    AutoScalePanel: TPanel;
    MarginsLabel: TLabel;
    AutoScaleRadioButton: TRadioButton;
    ExpressionLabel: TLabel;
    FontBitBtn: TBitBtn;
    CaptionLabel: TLabel;
    CaptionEdit: TEdit;
    LineWidthLabel: TLabel;
    LineWidthSpinEdit: TSpinEdit;
    TicksGroupBox: TGroupBox;
    MajorTicksLabel: TLabel;
    MajorTicksSpinEdit: TSpinEdit;
    MinorTicksLabel: TLabel;
    MinorTicksSpinEdit: TSpinEdit;
    LineColorGroupBox: TGroupBox;
    LineColorGrid: TColorGrid;
    LineColorBitBtn: TBitBtn;
    MarginsFloatEdit: TFloatEdit;
    MinScaleFloatEdit: TFloatEdit;
    MaxScaleFloatEdit: TFloatEdit;
    LabelsCheckBox: TCheckBox;
    VisibleCheckBox: TCheckBox;
    TicksRadioGroup: TRadioGroup;
    ExpressionComboBox: TExpressionComboBox;
    NumericFormatRadioButton: TRadioButton;
    DateTimeRadioButton: TRadioButton;
    NumericFormatPanel: TPanel;
    WidthLabel: TLabel;
    WidthSpinEdit: TSpinEdit;
    DigitsLabel: TLabel;
    DigitsSpinEdit: TSpinEdit;
    SampleLabel: TLabel;
    FormatRadioGroup: TRadioGroup;
    DateTimeFormatPanel: TPanel;
    DateTimeFormatLabel: TLabel;
    DateTimeFormatHelpLabel: TLabel;
    DateTimeFormatComboBox: TComboBox;
    SampleLabel1: TLabel;
    MinScaleDateTimePicker: TDateTimePicker;
    MaxScaleDateTimePicker: TDateTimePicker;
    HTMLTextBitBtn: TBitBtn;
    LinkedCheckBox: TCheckBox;
    ShowExpressionCheckBox: TCheckBox;
    InnerTicksCheckBox: TCheckBox;
    TickLengthLabel: TLabel;
    TickLengthFloatEdit: TFloatEdit;
    TicksCheckBox: TCheckBox;
    GridTabSheet: TTabSheet;
    GridCheckBox: TCheckBox;
    GridColorGroupBox: TGroupBox;
    GridColorGrid: TColorGrid;
    GridColorBitBtn: TBitBtn;
    GridWidthLabel: TLabel;
    GridWidthSpinEdit: TSpinEdit;
    GridLineTypeLabel: TLabel;
    GridLineTypeComboBox: TComboBox;
    GridColorDialog: TColorDialog;
    procedure LineColorGridChange(Sender: TObject);
    procedure LineColorBitBtnClick(Sender: TObject);
    procedure FontBitBtnClick(Sender: TObject);
    procedure FormatChange(Sender: TObject);
    procedure ScaleFloatEditChange(Sender: TObject);
    procedure LabelsCheckBoxClick(Sender: TObject);
    procedure TicksRadioGroupClick(Sender: TObject);
    procedure FormatRadioButtonClick(Sender: TObject);
    procedure DateTimeFormatHelpLabelMouseEnter(Sender: TObject);
    procedure DateTimeFormatHelpLabelMouseLeave(Sender: TObject);
    procedure DateTimeFormatComboBoxChange(Sender: TObject);
    procedure GridColorBitBtnClick(Sender: TObject);
    procedure GridColorGridChange(Sender: TObject);
    procedure ScaleDateTimePickerChange(Sender: TObject);
    procedure HTMLTextBitBtnClick(Sender: TObject);
    procedure GridLineTypeComboBoxDrawItem(Control: TWinControl; Index: Integer; Rect:
        TRect; State: TOwnerDrawState);
    procedure LinkedCheckBoxClick(Sender: TObject);
  private
    { Private declarations }
    FScaleIncrement: extended; // buffer set in Execute
    FPlotColorBuf: TColor; // used in HTMLTextBitBtnClick for preview painting
  public
    { Public declarations }
    function Execute(Axis: TAxis): boolean; // returns true if clicked OK
  end;

var 
  AxisPropsForm: TAxisPropsForm;

implementation

uses TextDlg, DateUtils, HTextDlg;

{$R *.DFM}

{ TAxisPropsForm }

function TAxisPropsForm.Execute(Axis: TAxis): boolean;
  function CheckScale: boolean; // true if scale is valid date/time
  var
    AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond: Word;
  begin
    Result:=false;
    DecodeDateTime(Axis.Min, AYear, AMonth, ADay, 
      AHour, AMinute, ASecond, AMilliSecond);
    if not IsValidDateTime(AYear, AMonth, ADay, 
      AHour, AMinute, ASecond, AMilliSecond)  
    then Exit;  
    DecodeDateTime(Axis.Max, AYear, AMonth, ADay, 
      AHour, AMinute, ASecond, AMilliSecond);
    if not IsValidDateTime(AYear, AMonth, ADay, 
      AHour, AMinute, ASecond, AMilliSecond)  
    then Exit;  
    Result:=true;
  end;
  function GetPickerDateTime(P: TDateTimePicker): extended; // QA0120062010
  var
    buf: array[0..100] of char;
    s: string;
    AYear, AMonth, ADay: Word;
  begin
    Result:=P.DateTime; // default; date changes are NOT applied!
    if GetWindowText(P.Handle, @buf, SizeOf(buf)-1)>0 then // while text is good...
    begin
      s:=buf; // implicit conversion
      if FormatDateTime(P.Format, Result)<>s then
      try // it seems that user changed date!
        ADay:=StrToInt(Copy(s, 13, 2)); // WARNING!!! depends on Format!!!
        AMonth:=StrToInt(Copy(s, 16, 2));
        AYear:=StrToInt(Copy(s, 19, 4));
        Result:=RecodeDate(Result, AYear, AMonth, ADay);
      except // just swallow exceptions by StrToInt and RecodeDate  
      end;
    end;
  end;
var 
  B: boolean; 
  I: integer;
begin
  // init FPlotColorBuf
  if Assigned(Axis.Plot)
  then FPlotColorBuf:=Axis.Plot.Color
  else FPlotColorBuf:=clBtnFace;
  // copy properties from axis
  MarginsFloatEdit.Value:=Axis.Margins; 
  MinScaleFloatEdit.Value:=Axis.Min;
  MaxScaleFloatEdit.Value:=Axis.Max;
  { note: if we would not check scale, DateTimePicker.DateTime:=Axis.Min
  on some values passed, but when DateTimePicker about to be displayed
  (in showmodal) it throws ECommonCalendarError which can't be processed.
  So in fact we try to invoke the exception where we can catch it. }
  try
    if CheckScale then
    begin
      MinScaleDateTimePicker.DateTime:=Axis.Min;
      MaxScaleDateTimePicker.DateTime:=Axis.Max;
    end else
      DateTimeFormatPanel.Font.Color:=clRed;
  except
    // Windows.Beep(1000, 200); for DEBUG only
    DateTimeFormatPanel.Font.Color:=clRed; 
  end;
  FScaleIncrement:=(Axis.Max-Axis.Min)/Axis.MajorTicks;
  if Axis.LogTicks then
  begin
    MinScaleFloatEdit.Multiply:=true;
    MaxScaleFloatEdit.Multiply:=true;
    MinScaleFloatEdit.Increment:=10;
    MaxScaleFloatEdit.Increment:=10;
  end else
  begin
    MinScaleFloatEdit.Multiply:=false;
    MaxScaleFloatEdit.Multiply:=false;
    MinScaleFloatEdit.Increment:=FScaleIncrement;
    MaxScaleFloatEdit.Increment:=FScaleIncrement;
  end;
  ExpressionComboBox.Text:=Axis.Expression;
  if Axis.SmartTicks // before format (see OnChange!)
  then TicksRadioGroup.ItemIndex:=1 else
    if Axis.LogTicks
    then TicksRadioGroup.ItemIndex:=2 
    else TicksRadioGroup.ItemIndex:=0;
  WidthSpinEdit.Value:=Axis.LabelWidth;
  DigitsSpinEdit.Value:=Axis.LabelDecimals;
  case Axis.labelType of
    ffGeneral: FormatRadioGroup.ItemIndex:=0;
    ffFixed: FormatRadioGroup.ItemIndex:=1;
    ffExponent: FormatRadioGroup.ItemIndex:=2;
  end;
  FormatChange(Self); // must be called AFTER setting format controls
  if Axis.LabelDateTime then
  begin
    DateTimeFormatComboBox.Text:=Axis.FormatString;
    DateTimeRadioButton.Checked:=true;
  end;  
  DateTimeFormatComboBoxChange(Self);
  AutoScaleRadioButton.Checked:=Axis.AutoScale;
  ManualScaleRadioButton.Checked:=not Axis.AutoScale; // else both may be off!
  CaptionEdit.Text:=Axis.Title;
  MajorTicksSpinEdit.Value:=Axis.MajorTicks;
  MinorTicksSpinEdit.Value:=Axis.MinorTicks;
  LineWidthSpinEdit.Value:=Axis.Pen.Width;
  GridCheckBox.Checked:=Axis.ShowGrid;
  LineColorGrid.ForegroundIndex:=-1;
  ColorDialog.Color:=Axis.Pen.Color;
  FontDialog.Font:=Axis.Font;
  TicksCheckBox.Checked:=Axis.TicksVisible;
  LabelsCheckBox.Checked:=Axis.LabelVisible;
  VisibleCheckBox.Checked:=Axis.Visible;
  ShowExpressionCheckBox.Checked:=Axis.ShowExpression;
  LabelsCheckBoxClick(nil);
  LinkedCheckBox.Checked:=Axis.IsLinked;
  InnerTicksCheckBox.Checked:=Axis.InnerTicks;
  TickLengthFloatEdit.Value:=Axis.TickLength+1;
  GridColorGrid.ForegroundIndex:=-1; // need for QA0127112009
  GridColorDialog.Color:=Axis.GridPen.Color;
  GridWidthSpinEdit.Value:=Axis.GridPen.Width;
  case Axis.GridPen.Style of
    psSolid: GridLineTypeComboBox.ItemIndex:=0;
    psDash: GridLineTypeComboBox.ItemIndex:=1;
    psDot: GridLineTypeComboBox.ItemIndex:=2;
    psDashDot: GridLineTypeComboBox.ItemIndex:=3;
    psDashDotDot: GridLineTypeComboBox.ItemIndex:=4;
    psClear: GridLineTypeComboBox.ItemIndex:=5;
  else GridLineTypeComboBox.ItemIndex:=-1;
  end;
  Result:=ShowModal=mrOK; 
  if not Result then Exit; // cancelled
  // copy properties to axis
  Axis.GridPen.Color:=GridColorDialog.Color;
  Axis.GridPen.Width:=GridWidthSpinEdit.Value;
  case GridLineTypeComboBox.ItemIndex of
    0: Axis.GridPen.Style:=psSolid;
    1: Axis.GridPen.Style:=psDash;
    2: Axis.GridPen.Style:=psDot;
    3: Axis.GridPen.Style:=psDashDot;
    4: Axis.GridPen.Style:=psDashDotDot;
    5: Axis.GridPen.Style:=psClear;
  end;
  Axis.TickLength:=TickLengthFloatEdit.Value-1;
  Axis.InnerTicks:=InnerTicksCheckBox.Checked;
  Axis.IsLinked:=LinkedCheckBox.Checked;
  Axis.ShowExpression:=ShowExpressionCheckBox.Checked;
  Axis.Margins:=MarginsFloatEdit.Value;   
  Axis.Expression:=ExpressionComboBox.Text;
  Axis.AutoScale:=AutoScaleRadioButton.Checked;
  Axis.LabelDateTime:=DateTimeRadioButton.Checked;
  if NumericFormatRadioButton.Checked then
  begin
    Axis.Min:=MinScaleFloatEdit.Value;
    Axis.Max:=MaxScaleFloatEdit.Value;
    Axis.LabelWidth:=WidthSpinEdit.Value;
    Axis.LabelDecimals:=DigitsSpinEdit.Value;
    case FormatRadioGroup.ItemIndex of
      0: Axis.labelType:=ffGeneral;
      1: Axis.labelType:=ffFixed;
      2: Axis.labelType:=ffExponent;
    end;
  end else
  begin
    Axis.Min:=GetPickerDateTime(MinScaleDateTimePicker);
    Axis.Max:=GetPickerDateTime(MaxScaleDateTimePicker);
    Axis.FormatString:=DateTimeFormatComboBox.Text;
  end;
  Axis.Title:=CaptionEdit.Text;
  Axis.MajorTicks:=MajorTicksSpinEdit.Value;
  Axis.MinorTicks:=MinorTicksSpinEdit.Value;
  Axis.Pen.Width:=LineWidthSpinEdit.Value;
  Axis.ShowGrid:=GridCheckBox.Checked;
  Axis.Pen.Color:=ColorDialog.Color;
  Axis.Font:=FontDialog.Font;
  Axis.TicksVisible:=TicksCheckBox.Checked;
  Axis.LabelVisible:=LabelsCheckBox.Checked;
  case TicksRadioGroup.ItemIndex of
    0: 
      begin
        Axis.LogTicks:=false; 
        Axis.SmartTicks:=false; 
      end;
    1:
      Axis.SmartTicks:=true;
    2:
      Axis.LogTicks:=true;    
  end; 
  Axis.Visible:=VisibleCheckBox.Checked;
  if not ExpressionComboBox.EditorModified
  then Exit; // QA0119082009
  B:=true; // check expression for duplication and add to the top of list
  with ExpressionComboBox.Items do 
  for I:=0 to Count-1 do
  if Strings[I]=ExpressionComboBox.Text 
  then B:=false;
  if B and (ExpressionComboBox.Text<>'')
  then ExpressionComboBox.Items.Insert(0, ExpressionComboBox.Text);
end;

procedure TAxisPropsForm.LineColorGridChange(Sender: TObject);
begin
  if LineColorGrid.Tag=0 then // bugfix!
  ColorDialog.Color:=LineColorGrid.ForegroundColor; 
end;

procedure TAxisPropsForm.LineColorBitBtnClick(Sender: TObject);
begin 
  if ColorDialog.Execute then
  try
    LineColorGrid.Tag:=1; // prevent changing color in LineColorGridChange
    LineColorGrid.ForegroundIndex:=-1; 
  finally
    LineColorGrid.Tag:=0;
  end;  
end;

procedure TAxisPropsForm.FontBitBtnClick(Sender: TObject);
begin 
  FontDialog.Execute; 
end;

procedure TAxisPropsForm.FormatChange(Sender: TObject);
var
  F: TFloatFormat; 
begin
  (*
  case FormatRadioGroup.ItemIndex of // update scale editors
    0: MinScaleFloatEdit.FType:=ffGeneral;
    1: MinScaleFloatEdit.FType:=ffFixed;
    2: MinScaleFloatEdit.FType:=ffExponent;
  end;{case}
  MinScaleFloatEdit.MinWidth:=WidthSpinEdit.Value;
  MinScaleFloatEdit.Decimals:=DigitsSpinEdit.Value;
  MaxScaleFloatEdit.FType:=MinScaleFloatEdit.FType;
  MaxScaleFloatEdit.MinWidth:=MinScaleFloatEdit.MinWidth;
  MaxScaleFloatEdit.Decimals:=MinScaleFloatEdit.Decimals;
  with MaxScaleFloatEdit do 
  SampleLabel.Caption:=FloatToStrF((Value+
    MinScaleFloatEdit.Value)/2,FType,MinWidth,Decimals); // update sample
  Because changing format may cause parasitic rounding, we will NOT use it!
  *)
  case FormatRadioGroup.ItemIndex of 
    0: F:=ffGeneral;
    1: F:=ffFixed;
    2: F:=ffExponent;
  else
    F:=ffGeneral;  
  end;
  SampleLabel.Caption:=FloatToStrF((MinScaleFloatEdit.Value+
    MaxScaleFloatEdit.Value)/2, F, WidthSpinEdit.Value, DigitsSpinEdit.Value);
  if F=ffFixed // correct to avoid rounding
  then F:=ffGeneral;
  MinScaleFloatEdit.FType:=F;
  MaxScaleFloatEdit.FType:=F;
end;

procedure TAxisPropsForm.ScaleFloatEditChange(Sender: TObject);
begin
  if (Sender as TFloatEdit).Focused 
  then ManualScaleRadioButton.Checked:=true;
end;

procedure TAxisPropsForm.LabelsCheckBoxClick(Sender: TObject);
begin
  TicksCheckBox.Enabled:=not LabelsCheckBox.Checked;
  if LabelsCheckBox.Checked
  then TicksCheckBox.Checked:=true;
  GridCheckBox.Enabled:=(LabelsCheckBox.Checked or TicksCheckBox.Checked) 
    and (not LinkedCheckBox.Checked); // QA0107012010
  if not GridCheckBox.Enabled 
  then GridCheckBox.Checked:=false;
end;

procedure TAxisPropsForm.TicksRadioGroupClick(Sender: TObject);
begin
  if TicksRadioGroup.ItemIndex=2 then
  begin
    MinorTicksSpinEdit.Value:=9;
    MinScaleFloatEdit.Multiply:=true;
    MaxScaleFloatEdit.Multiply:=true;
    MinScaleFloatEdit.Increment:=10;
    MaxScaleFloatEdit.Increment:=10;
  end else 
  begin
    MinScaleFloatEdit.Multiply:=false;
    MaxScaleFloatEdit.Multiply:=false;
    MinScaleFloatEdit.Increment:=FScaleIncrement;
    MaxScaleFloatEdit.Increment:=FScaleIncrement;
    MinorTicksSpinEdit.Value:=2;
  end;  
end;

procedure TAxisPropsForm.FormatRadioButtonClick(Sender: TObject);
begin
  NumericFormatPanel.Visible:=NumericFormatRadioButton.Checked;
  DateTimeFormatPanel.Visible:=DateTimeRadioButton.Checked;
  MinScaleDateTimePicker.Visible:=DateTimeRadioButton.Checked;
  MaxScaleDateTimePicker.Visible:=DateTimeRadioButton.Checked;
  MinScaleFloatEdit.Visible:=NumericFormatRadioButton.Checked;
  MaxScaleFloatEdit.Visible:=NumericFormatRadioButton.Checked;
end;

procedure TAxisPropsForm.DateTimeFormatHelpLabelMouseEnter(Sender: TObject);
begin
  DateTimeFormatHelpLabel.Font.Color:=clBlue;
end;

procedure TAxisPropsForm.DateTimeFormatHelpLabelMouseLeave(Sender: TObject);
begin
  DateTimeFormatHelpLabel.Font.Color:=clNavy;
end;

procedure TAxisPropsForm.DateTimeFormatComboBoxChange(Sender: TObject);
begin
  SampleLabel1.Caption:=FormatDateTime(DateTimeFormatComboBox.Text,
    (MinScaleDateTimePicker.DateTime+MaxScaleDateTimePicker.DateTime)/2);
end; // note: we use ^ DateTimePickers, not FloatEdits!

procedure TAxisPropsForm.GridColorBitBtnClick(Sender: TObject);
begin
  if GridColorDialog.Execute then
  try
    GridColorGrid.Tag:=1; // prevent changing color in LineColorGridChange
    GridColorGrid.ForegroundIndex:=-1; 
  finally
    GridColorGrid.Tag:=0;
  end;  
end;

procedure TAxisPropsForm.GridColorGridChange(Sender: TObject);
begin
  if GridColorGrid.Tag=0 // bugfix!
  then GridColorDialog.Color:=GridColorGrid.ForegroundColor; 
end;

procedure TAxisPropsForm.ScaleDateTimePickerChange(Sender: TObject);
begin
  if (Sender as TDateTimePicker).Focused 
  then ManualScaleRadioButton.Checked:=true;
end;

procedure TAxisPropsForm.HTMLTextBitBtnClick(Sender: TObject);
var
  S: string;
begin
  S:=CaptionEdit.Text;
  HTMLTextForm:=THTMLTextForm.Create(Application);
  try
    HTMLTextForm.HelpBitBtn.OnClick:=HelpBitBtn.OnClick;
    if HTMLTextForm.Execute(S, FPlotColorBuf, FontDialog.Font)
    then CaptionEdit.Text:=S;
  finally
    HTMLTextForm.Free;
  end;
end;

// copypasted from SerieDlg.TSeriePropsForm.LineTypeComboBoxDrawItem
procedure TAxisPropsForm.GridLineTypeComboBoxDrawItem(Control: TWinControl; Index:
    Integer; Rect: TRect; State: TOwnerDrawState);
var 
  PSBuf: TPenStyle;
  ISize, PSize: integer;
  CBuf: TColor;
begin
  with GridLineTypeComboBox do
  begin
    Canvas.FillRect(Rect); 
    PSBuf:=Canvas.Pen.Style;
    case Index of
      0,5: Canvas.Pen.Style:=psSolid;
      1: Canvas.Pen.Style:=psDash;
      2: Canvas.Pen.Style:=psDot;
      3: Canvas.Pen.Style:=psDashDot;
      4: Canvas.Pen.Style:=psDashDotDot;
    end;
    if Index<>5 then
    begin
      Canvas.MoveTo(Rect.Left, Rect.Top+(Rect.Bottom-Rect.Top) div 2);
      Canvas.LineTo(Rect.Right, Rect.Top+(Rect.Bottom-Rect.Top) div 2);
    end else
    begin
      CBuf:=Canvas.Pen.Color; 
      PSize:=Canvas.Pen.Width;
      Canvas.Pen.Color:=clRed;
      Canvas.Pen.Width:=2;
      ISize:={LineTypeComboBox.Height} (Rect.Bottom-Rect.Top) div 3;
      Canvas.MoveTo(Rect.Left+ISize, Rect.Top+((Rect.Bottom-Rect.Top) div 2)-ISize);
      Canvas.LineTo(Rect.Left+ISize*3, Rect.Top+((Rect.Bottom-Rect.Top) div 2)+ISize);
      Canvas.MoveTo(Rect.Left+ISize*3, Rect.Top+((Rect.Bottom-Rect.Top) div 2)-ISize);
      Canvas.LineTo(Rect.Left+ISize, Rect.Top+((Rect.Bottom-Rect.Top) div 2)+ISize);
      Canvas.Pen.Color:=CBuf;
      Canvas.Pen.Width:=PSize;
      Canvas.TextOut(Rect.Left+ISize*4, Rect.Top, Items[Index]);    
    end;
    Canvas.Pen.Style:=PSBuf;
  end;
end;

procedure TAxisPropsForm.LinkedCheckBoxClick(Sender: TObject);
begin
  AutoScaleRadioButton.Enabled:=not LinkedCheckBox.Checked;
  AutoScalePanel.Enabled:=not LinkedCheckBox.Checked;
  ManualScaleRadioButton.Enabled:=not LinkedCheckBox.Checked;
  ManualScalePanel.Enabled:=not LinkedCheckBox.Checked;
  TicksGroupBox.Enabled:=not LinkedCheckBox.Checked;
  TicksRadioGroup.Enabled:=not LinkedCheckBox.Checked;
  GridCheckBox.Enabled:=not LinkedCheckBox.Checked;
  if LinkedCheckBox.Checked
  then GridCheckBox.Checked:=false;  
end;

end.
 